@@ -28,20 +28,23 @@ from unicodedata import normalize
 import hashlib
 import re
 from typing import Tuple, TYPE_CHECKING, Union, Sequence, Optional, Dict, List, NamedTuple
-from functools import lru_cache
+from functools import lru_cache, wraps
 from abc import ABC, abstractmethod
 
 from . import bitcoin, ecc, constants, bip32
 from .bitcoin import deserialize_privkey, serialize_privkey, BaseDecodeError
 from .transaction import Transaction, PartialTransaction, PartialTxInput, PartialTxOutput, TxInput
-from .bip32 import (convert_bip32_path_to_list_of_uint32, BIP32_PRIME,
+from .bip32 import (convert_bip32_strpath_to_intpath, BIP32_PRIME,
                     is_xpub, is_xprv, BIP32Node, normalize_bip32_derivation,
-                    convert_bip32_intpath_to_strpath, is_xkey_consistent_with_key_origin_info)
+                    convert_bip32_intpath_to_strpath, is_xkey_consistent_with_key_origin_info,
+                    KeyOriginInfo)
+from .descriptor import PubkeyProvider
 from .ecc import string_to_number
 from .crypto import (pw_decode, pw_encode, sha256, sha256d, PW_HASH_VERSION_LATEST,
-                     SUPPORTED_PW_HASH_VERSIONS, UnsupportedPasswordHashVersion, hash_160)
+                     SUPPORTED_PW_HASH_VERSIONS, UnsupportedPasswordHashVersion, hash_160,
+                     CiphertextFormatError)
 from .util import (InvalidPassword, WalletFileException,
-                   BitcoinException, bh2u, bfh, inv_dict, is_hex_str)
+                   BitcoinException, bfh, inv_dict, is_hex_str)
 from .mnemonic import Mnemonic, Wordlist, seed_type, is_seed
 from .plugin import run_hook
 from .logging import Logger
@@ -50,11 +53,33 @@ if TYPE_CHECKING:
     from .gui.qt.util import TaskThread
     from .plugins.hw_wallet import HW_PluginBase, HardwareClientBase, HardwareHandlerBase
     from .wallet_db import WalletDB
+    from .plugin import Device
 
 
 class CannotDerivePubkey(Exception): pass
 
 
+def also_test_none_password(check_password_fn):
+    """Decorator for check_password, simply to give a friendlier exception if
+    check_password(x) is called on a keystore that does not have a password set.
+    """
+    @wraps(check_password_fn)
+    def wrapper(self: 'Software_KeyStore', *args):
+        password = args[0]
+        try:
+            return check_password_fn(self, password)
+        except (CiphertextFormatError, InvalidPassword) as e:
+            if password is not None:
+                try:
+                    check_password_fn(self, None)
+                except Exception:
+                    pass
+                else:
+                    raise InvalidPassword("password given but keystore has no password") from e
+            raise
+    return wrapper
+
+
 class KeyStore(Logger, ABC):
     type: str
 
@@ -127,7 +152,14 @@ class KeyStore(Logger, ABC):
         pass
 
     @abstractmethod
-    def sign_message(self, sequence: 'AddressIndexGeneric', message, password) -> bytes:
+    def sign_message(
+            self,
+            sequence: 'AddressIndexGeneric',
+            message: str,
+            password,
+            *,
+            script_type: Optional[str] = None,
+    ) -> bytes:
         pass
 
     @abstractmethod
@@ -149,6 +181,10 @@ class KeyStore(Logger, ABC):
         """
         pass
 
+    @abstractmethod
+    def get_pubkey_provider(self, sequence: 'AddressIndexGeneric') -> Optional[PubkeyProvider]:
+        pass
+
     def find_my_pubkey_in_txinout(
             self, txinout: Union['PartialTxInput', 'PartialTxOutput'],
             *, only_der_suffix: bool = False
@@ -175,7 +211,7 @@ class Software_KeyStore(KeyStore):
     def may_have_password(self):
         return not self.is_watching_only()
 
-    def sign_message(self, sequence, message, password) -> bytes:
+    def sign_message(self, sequence, message, password, *, script_type=None) -> bytes:
         privkey, compressed = self.get_private_key(sequence, password)
         key = ecc.ECPrivkey(privkey)
         return key.sign_message(message, compressed)
@@ -204,7 +240,8 @@ class Software_KeyStore(KeyStore):
         pass
 
     @abstractmethod
-    def check_password(self, password):
+    def check_password(self, password: Optional[str]) -> None:
+        """Raises InvalidPassword if password is not correct"""
         pass
 
     @abstractmethod
@@ -235,6 +272,7 @@ class Imported_KeyStore(Software_KeyStore):
     def can_import(self):
         return True
 
+    @also_test_none_password
     def check_password(self, password):
         pubkey = list(self.keypairs.keys())[0]
         self.get_private_key(pubkey, password)
@@ -270,6 +308,15 @@ class Imported_KeyStore(Software_KeyStore):
             return pubkey.hex()
         return None
 
+    def get_pubkey_provider(self, sequence: 'AddressIndexGeneric') -> Optional[PubkeyProvider]:
+        if sequence in self.keypairs:
+            return PubkeyProvider(
+                origin=None,
+                pubkey=sequence,
+                deriv_path=None,
+            )
+        return None
+
     def update_password(self, old_password, new_password):
         self.check_password(old_password)
         if new_password == '':
@@ -371,6 +418,9 @@ class MasterPublicKeyMixin(ABC):
         """
         pass
 
+    def get_key_origin_info(self) -> Optional[KeyOriginInfo]:
+        return None
+
     @abstractmethod
     def derive_pubkey(self, for_change: int, n: int) -> bytes:
         """Returns pubkey at given path.
@@ -404,7 +454,7 @@ class MasterPublicKeyMixin(ABC):
         # 1. try fp against our root
         ks_root_fingerprint_hex = self.get_root_fingerprint()
         ks_der_prefix_str = self.get_derivation_prefix()
-        ks_der_prefix = convert_bip32_path_to_list_of_uint32(ks_der_prefix_str) if ks_der_prefix_str else None
+        ks_der_prefix = convert_bip32_strpath_to_intpath(ks_der_prefix_str) if ks_der_prefix_str else None
         if (ks_root_fingerprint_hex is not None and ks_der_prefix is not None and
                 fp_found.hex() == ks_root_fingerprint_hex):
             if path_found[:len(ks_der_prefix)] == ks_der_prefix:
@@ -474,11 +524,11 @@ class Xpub(MasterPublicKeyMixin):
         if not only_der_suffix and fingerprint_hex is not None and der_prefix_str is not None:
             # use root fp, and true full path
             fingerprint_bytes = bfh(fingerprint_hex)
-            der_prefix_ints = convert_bip32_path_to_list_of_uint32(der_prefix_str)
+            der_prefix_ints = convert_bip32_strpath_to_intpath(der_prefix_str)
         else:
             # use intermediate fp, and claim der suffix is the full path
             fingerprint_bytes = self.get_bip32_node_for_xpub().calc_fingerprint_of_this_node()
-            der_prefix_ints = convert_bip32_path_to_list_of_uint32('m')
+            der_prefix_ints = convert_bip32_strpath_to_intpath('m')
         der_full = der_prefix_ints + list(der_suffix)
         return fingerprint_bytes, der_full
 
@@ -491,11 +541,31 @@ class Xpub(MasterPublicKeyMixin):
         child_number_int = der_full[-1] if len(der_full) >= 1 else 0
         child_number_bytes = child_number_int.to_bytes(length=4, byteorder="big")
         fingerprint = bytes(4) if depth == 0 else bip32node.fingerprint
-        bip32node = bip32node._replace(depth=depth,
-                                       fingerprint=fingerprint,
-                                       child_number=child_number_bytes)
+        bip32node = bip32node._replace(
+            depth=depth,
+            fingerprint=fingerprint,
+            child_number=child_number_bytes,
+            # only put plain xpubs (not ypub/zpub) in PSBTs:
+            xtype="standard",
+        )
         return bip32node.to_xpub()
 
+    def get_key_origin_info(self) -> Optional[KeyOriginInfo]:
+        fp_bytes, der_full = self.get_fp_and_derivation_to_be_used_in_partial_tx(
+            der_suffix=[], only_der_suffix=False)
+        origin = KeyOriginInfo(fingerprint=fp_bytes, path=der_full)
+        return origin
+
+    def get_pubkey_provider(self, sequence: 'AddressIndexGeneric') -> Optional[PubkeyProvider]:
+        strpath = convert_bip32_intpath_to_strpath(sequence)
+        strpath = strpath[1:]  # cut leading "m"
+        bip32node = self.get_bip32_node_for_xpub()
+        return PubkeyProvider(
+            origin=self.get_key_origin_info(),
+            pubkey=bip32node._replace(xtype="standard").to_xkey(),
+            deriv_path=strpath,
+        )
+
     def add_key_origin_from_root_node(self, *, derivation_prefix: str, root_node: BIP32Node):
         assert self.xpub
         # try to derive ourselves from what we were given
@@ -568,6 +638,7 @@ class BIP32_KeyStore(Xpub, Deterministic_KeyStore):
     def get_master_private_key(self, password):
         return pw_decode(self.xprv, password, version=self.pw_hash_version)
 
+    @also_test_none_password
     def check_password(self, password):
         xprv = pw_decode(self.xprv, password, version=self.pw_hash_version)
         try:
@@ -734,6 +805,7 @@ class Old_KeyStore(MasterPublicKeyMixin, Deterministic_KeyStore):
         if master_public_key != bfh(self.mpk):
             raise InvalidPassword()
 
+    @also_test_none_password
     def check_password(self, password):
         seed = self.get_hex_seed(password)
         self._check_seed(seed)
@@ -760,10 +832,17 @@ class Old_KeyStore(MasterPublicKeyMixin, Deterministic_KeyStore):
         fingerprint_hex = self.get_root_fingerprint()
         der_prefix_str = self.get_derivation_prefix()
         fingerprint_bytes = bfh(fingerprint_hex)
-        der_prefix_ints = convert_bip32_path_to_list_of_uint32(der_prefix_str)
+        der_prefix_ints = convert_bip32_strpath_to_intpath(der_prefix_str)
         der_full = der_prefix_ints + list(der_suffix)
         return fingerprint_bytes, der_full
 
+    def get_pubkey_provider(self, sequence: 'AddressIndexGeneric') -> Optional[PubkeyProvider]:
+        return PubkeyProvider(
+            origin=None,
+            pubkey=self.derive_pubkey(*sequence).hex(),
+            deriv_path=None,
+        )
+
     def update_password(self, old_password, new_password):
         self.check_password(old_password)
         if new_password == '':
@@ -789,7 +868,7 @@ class Hardware_KeyStore(Xpub, KeyStore):
         # handler.  The handler is per-window and preserved across
         # device reconnects
         self.xpub = d.get('xpub')
-        self.label = d.get('label')
+        self.label = d.get('label')  # type: Optional[str]
         self.soft_device_id = d.get('soft_device_id')  # type: Optional[str]
         self.handler = None  # type: Optional[HardwareHandlerBase]
         run_hook('init_keystore', self)
@@ -833,14 +912,31 @@ class Hardware_KeyStore(Xpub, KeyStore):
         assert not self.has_seed()
         return False
 
+    def get_client(
+            self,
+            force_pair: bool = True,
+            *,
+            devices: Sequence['Device'] = None,
+            allow_user_interaction: bool = True,
+    ) -> Optional['HardwareClientBase']:
+        return self.plugin.get_client(
+            self,
+            force_pair=force_pair,
+            devices=devices,
+            allow_user_interaction=allow_user_interaction,
+        )
+
     def get_password_for_storage_encryption(self) -> str:
-        client = self.plugin.get_client(self)
+        client = self.get_client()
         return client.get_password_for_storage_encryption()
 
     def has_usable_connection_with_device(self) -> bool:
-        if not hasattr(self, 'plugin'):
-            return False
-        client = self.plugin.get_client(self, force_pair=False)
+        # we try to create a client even if there isn't one already,
+        # but do not prompt the user if auto-select fails:
+        client = self.get_client(
+            force_pair=True,
+            allow_user_interaction=False,
+        )
         if client is None:
             return False
         return client.has_usable_connection_with_device()
@@ -860,6 +956,12 @@ class Hardware_KeyStore(Xpub, KeyStore):
             self.soft_device_id = client.get_soft_device_id()
             self.is_requesting_to_be_rewritten_to_wallet_file = True
 
+    def pairing_code(self) -> Optional[str]:
+        """Used by the DeviceMgr to keep track of paired hw devices."""
+        if not self.soft_device_id:
+            return None
+        return f"{self.plugin.name}/{self.soft_device_id}"
+
 
 KeyStoreWithMPK = Union[KeyStore, MasterPublicKeyMixin]  # intersection really...
 AddressIndexGeneric = Union[Sequence[int], str]  # can be hex pubkey str
@@ -928,7 +1030,7 @@ PURPOSE48_SCRIPT_TYPES_INV = inv_dict(PURPOSE48_SCRIPT_TYPES)
 
 def xtype_from_derivation(derivation: str) -> str:
     """Returns the script type to be used for this derivation."""
-    bip32_indices = convert_bip32_path_to_list_of_uint32(derivation)
+    bip32_indices = convert_bip32_strpath_to_intpath(derivation)
     if len(bip32_indices) >= 1:
         if bip32_indices[0] == 84 + BIP32_PRIME:
             return 'p2wpkh'
