@@ -31,23 +31,23 @@ import threading
 import socket
 import json
 import sys
-import asyncio
-from typing import NamedTuple, Optional, Sequence, List, Dict, Tuple, TYPE_CHECKING, Iterable, Set, Any
+from typing import NamedTuple, Optional, Sequence, List, Dict, Tuple, TYPE_CHECKING, Iterable, Set, Any, TypeVar
 import traceback
 import concurrent
 from concurrent import futures
 import copy
 import functools
+from enum import IntEnum
 
 import aiorpcx
-from aiorpcx import TaskGroup, ignore_after
+from aiorpcx import ignore_after
 from aiohttp import ClientResponse
 
 from . import util
-from .util import (log_exceptions, ignore_exceptions,
-                   bfh, SilentTaskGroup, make_aiohttp_session, send_exception_to_crash_reporter,
+from .util import (log_exceptions, ignore_exceptions, OldTaskGroup,
+                   bfh, make_aiohttp_session, send_exception_to_crash_reporter,
                    is_hash256_str, is_non_negative_integer, MyEncoder, NetworkRetryManager,
-                   nullcontext)
+                   nullcontext, error_text_str_to_safe_str)
 from .bitcoin import COIN
 from . import constants
 from . import blockchain
@@ -62,10 +62,12 @@ from .version import PROTOCOL_VERSION
 from .simple_config import SimpleConfig
 from .i18n import _
 from .logging import get_logger, Logger
-from .lnutil import ChannelBlackList
 
 if TYPE_CHECKING:
+    from collections.abc import Coroutine
+
     from .channel_db import ChannelDB
+    from .lnrouter import LNPathFinder
     from .lnworker import LNGossip
     from .lnwatcher import WatchTower
     from .daemon import Daemon
@@ -78,6 +80,14 @@ NUM_TARGET_CONNECTED_SERVERS = 10
 NUM_STICKY_SERVERS = 4
 NUM_RECENT_SERVERS = 20
 
+T = TypeVar('T')
+
+
+class ConnectionState(IntEnum):
+    DISCONNECTED  = 0
+    CONNECTING    = 1
+    CONNECTED     = 2
+
 
 def parse_servers(result: Sequence[Tuple[str, str, List[str]]]) -> Dict[str, dict]:
     """Convert servers list (from protocol method "server.peers.subscribe") into dict format.
@@ -232,8 +242,9 @@ class UntrustedServerReturnedError(NetworkException):
         return _("The server returned an error.")
 
     def __repr__(self):
+        e = self.original_exception
         return (f"<UntrustedServerReturnedError "
-                f"[DO NOT TRUST THIS MESSAGE] original_exception: {repr(self.original_exception)}>")
+                f"[DO NOT TRUST THIS MESSAGE] original_exception: {error_text_str_to_safe_str(repr(e))}>")
 
 
 _INSTANCE = None
@@ -246,7 +257,7 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
 
     LOGGING_SHORTCUT = 'n'
 
-    taskgroup: Optional[TaskGroup]
+    taskgroup: Optional[OldTaskGroup]
     interface: Optional[Interface]
     interfaces: Dict[ServerAddr, Interface]
     _connecting_ifaces: Set[ServerAddr]
@@ -254,10 +265,10 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
     default_server: ServerAddr
     _recent_servers: List[ServerAddr]
 
-    channel_blacklist: 'ChannelBlackList'
     channel_db: Optional['ChannelDB'] = None
     lngossip: Optional['LNGossip'] = None
     local_watchtower: Optional['WatchTower'] = None
+    path_finder: Optional['LNPathFinder'] = None
 
     def __init__(self, config: SimpleConfig, *, daemon: 'Daemon' = None):
         global _INSTANCE
@@ -273,13 +284,8 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
             init_retry_delay_urgent=1,
         )
 
-        self.asyncio_loop = asyncio.get_event_loop()
+        self.asyncio_loop = util.get_asyncio_loop()
         assert self.asyncio_loop.is_running(), "event loop not running"
-        try:
-            self._loop_thread = self.asyncio_loop._mythread  # type: threading.Thread  # only used for sanity checks
-        except AttributeError as e:
-            self.logger.warning(f"asyncio loop does not have _mythread set: {e!r}")
-            self._loop_thread = None
 
         assert isinstance(config, SimpleConfig), f"config should be a SimpleConfig instead of {type(config)}"
         self.config = config
@@ -296,18 +302,7 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
 
         self._allowed_protocols = {PREFERRED_NETWORK_PROTOCOL}
 
-        # Server for addresses and transactions
-        self.default_server = self.config.get('server', None)
-        # Sanitize default server
-        if self.default_server:
-            try:
-                self.default_server = ServerAddr.from_str(self.default_server)
-            except:
-                self.logger.warning('failed to parse server-string; falling back to localhost:1:s.')
-                self.default_server = ServerAddr.from_str("localhost:1:s")
-        else:
-            self.default_server = pick_random_server(allowed_protocols=self._allowed_protocols)
-        assert isinstance(self.default_server, ServerAddr), f"invalid type for default_server: {self.default_server!r}"
+        self._init_parameters_from_config()
 
         self.taskgroup = None
 
@@ -339,22 +334,17 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
         self.interfaces = {}  # these are the ifaces in "initialised and usable" state
         self._closing_ifaces = set()
 
-        self.auto_connect = self.config.get('auto_connect', True)
-        self.proxy = None
-        self._maybe_set_oneserver()
-
         # Dump network messages (all interfaces).  Set at runtime from the console.
         self.debug = False
 
-        self._set_status('disconnected')
+        self._set_status(ConnectionState.DISCONNECTED)
         self._has_ever_managed_to_connect_to_server = False
+        self._was_started = False
 
         # lightning network
-        self.channel_blacklist = ChannelBlackList()
         if self.config.get('run_watchtower', False):
             from . import lnwatcher
             self.local_watchtower = lnwatcher.WatchTower(self)
-            self.local_watchtower.start_network(self)
             asyncio.ensure_future(self.local_watchtower.start_watching())
 
     def has_internet_connection(self) -> bool:
@@ -385,10 +375,13 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
             if full_shutdown:
                 await self.channel_db.stopped_event.wait()
             self.channel_db = None
+            self.path_finder = None
 
-    def run_from_another_thread(self, coro, *, timeout=None):
-        assert self._loop_thread != threading.current_thread(), 'must not be called from network thread'
-        fut = asyncio.run_coroutine_threadsafe(coro, self.asyncio_loop)
+    @classmethod
+    def run_from_another_thread(cls, coro: 'Coroutine[Any, Any, T]', *, timeout=None) -> T:
+        loop = util.get_asyncio_loop()
+        assert util.get_running_loop() != loop, 'must not be called from asyncio thread'
+        fut = asyncio.run_coroutine_threadsafe(coro, loop)
         return fut.result(timeout)
 
     @staticmethod
@@ -438,14 +431,22 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
 
     def _set_status(self, status):
         self.connection_status = status
-        self.notify('status')
+        util.trigger_callback('status')
 
     def is_connected(self):
         interface = self.interface
-        return interface is not None and interface.ready.done()
+        return interface is not None and interface.is_connected_and_ready()
 
     def is_connecting(self):
-        return self.connection_status == 'connecting'
+        return self.connection_status == ConnectionState.CONNECTING
+
+    def get_connection_status_for_GUI(self):
+        ConnectionStates = {
+            ConnectionState.DISCONNECTED: _('Disconnected'),
+            ConnectionState.CONNECTING: _('Connecting'),
+            ConnectionState.CONNECTED: _('Connected'),
+        }
+        return ConnectionStates[self.connection_status]
 
     async def _request_server_info(self, interface: 'Interface'):
         await interface.ready
@@ -453,7 +454,7 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
 
         async def get_banner():
             self.banner = await interface.get_server_banner()
-            self.notify('banner')
+            util.trigger_callback('banner', self.banner)
         async def get_donation_address():
             self.donation_address = await interface.get_donation_address()
         async def get_server_peers():
@@ -463,11 +464,11 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
             server_peers = server_peers[:max_accepted_peers]
             # note that 'parse_servers' also validates the data (which is untrusted input!)
             self.server_peers = parse_servers(server_peers)
-            self.notify('servers')
+            util.trigger_callback('servers', self.get_servers())
         async def get_relay_fee():
             self.relay_fee = await interface.get_relay_fee()
 
-        async with TaskGroup() as group:
+        async with OldTaskGroup() as group:
             await group.spawn(get_banner)
             await group.spawn(get_donation_address)
             await group.spawn(get_server_peers)
@@ -479,28 +480,7 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
         histogram = await interface.get_fee_histogram()
         self.config.mempool_fees = histogram
         self.logger.info(f'fee_histogram {histogram}')
-        self.notify('fee_histogram')
-
-    def get_status_value(self, key):
-        if key == 'status':
-            value = self.connection_status
-        elif key == 'banner':
-            value = self.banner
-        elif key == 'fee':
-            value = self.config.fee_estimates
-        elif key == 'fee_histogram':
-            value = self.config.mempool_fees
-        elif key == 'servers':
-            value = self.get_servers()
-        else:
-            raise Exception('unexpected trigger key {}'.format(key))
-        return value
-
-    def notify(self, key):
-        if key in ['status', 'updated']:
-            util.trigger_callback(key)
-        else:
-            util.trigger_callback(key, self.get_status_value(key))
+        util.trigger_callback('fee_histogram', self.config.mempool_fees)
 
     def get_parameters(self) -> NetworkParameters:
         return NetworkParameters(server=self.default_server,
@@ -508,6 +488,12 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
                                  auto_connect=self.auto_connect,
                                  oneserver=self.oneserver)
 
+    def _init_parameters_from_config(self) -> None:
+        self.auto_connect = self.config.get('auto_connect', True)
+        self._set_default_server()
+        self._set_proxy(deserialize_proxy(self.config.get('proxy')))
+        self._maybe_set_oneserver()
+
     def get_donation_address(self):
         if self.is_connected():
             return self.donation_address
@@ -517,6 +503,10 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
         with self.interfaces_lock:
             return list(self.interfaces)
 
+    def get_status(self):
+        n = len(self.get_interfaces())
+        return _("Connected to {0} nodes.").format(n) if n > 1 else _("Connected to {0} node.").format(n) if n == 1 else _("Not connected")
+
     def get_fee_estimates(self):
         from statistics import median
         from .simple_config import FEE_ETA_TARGETS
@@ -534,7 +524,7 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
                 return {}
             return self.interface.fee_estimates_eta
 
-    def update_fee_estimates(self, *, fee_est: Dict = None):
+    def update_fee_estimates(self, *, fee_est: Dict[int, int] = None):
         if fee_est is None:
             fee_est = self.get_fee_estimates()
         for nblock_target, fee in fee_est.items():
@@ -542,7 +532,7 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
         if not hasattr(self, "_prev_fee_est") or self._prev_fee_est != fee_est:
             self._prev_fee_est = copy.copy(fee_est)
             self.logger.info(f'fee_estimates {fee_est}')
-        self.notify('fee')
+        util.trigger_callback('fee', self.config.fee_estimates)
 
     @with_recent_servers_lock
     def get_servers(self):
@@ -598,11 +588,33 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
             return server
         return None
 
+    def _set_default_server(self) -> None:
+        # Server for addresses and transactions
+        server = self.config.get('server', None)
+        # Sanitize default server
+        if server:
+            try:
+                self.default_server = ServerAddr.from_str(server)
+            except:
+                self.logger.warning(f'failed to parse server-string ({server!r}); falling back to localhost:1:s.')
+                self.default_server = ServerAddr.from_str("localhost:1:s")
+        else:
+            self.default_server = pick_random_server(allowed_protocols=self._allowed_protocols)
+        assert isinstance(self.default_server, ServerAddr), f"invalid type for default_server: {self.default_server!r}"
+
     def _set_proxy(self, proxy: Optional[dict]):
         self.proxy = proxy
         dns_hacks.configure_dns_depending_on_proxy(bool(proxy))
         self.logger.info(f'setting proxy {proxy}')
-        util.trigger_callback('proxy_set', self.proxy)
+
+        self.tor_proxy = False
+        if bool(proxy) and proxy['mode'] == 'socks5':
+            # test for Tor
+            self.tor_proxy = util.is_tor_socks_port(proxy['host'], int(proxy['port']))
+            if self.tor_proxy:
+                self.logger.info(f'Proxy is TOR')
+
+        util.trigger_callback('proxy_set', self.proxy, self.tor_proxy)
 
     @log_exceptions
     async def set_parameters(self, net_params: NetworkParameters):
@@ -626,14 +638,19 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
                 or self.config.get('oneserver') != net_params.oneserver:
             return
 
+        proxy_changed = self.proxy != proxy
+        oneserver_changed = self.oneserver != net_params.oneserver
+        default_server_changed = self.default_server != server
+        self._init_parameters_from_config()
+        if not self._was_started:
+            return
+
         async with self.restart_lock:
-            self.auto_connect = net_params.auto_connect
-            if self.proxy != proxy or self.oneserver != net_params.oneserver:
-                # Restart the network defaulting to the given server
+            if proxy_changed or oneserver_changed:
+                # Restart the network
                 await self.stop(full_shutdown=False)
-                self.default_server = server
                 await self._start()
-            elif self.default_server != server:
+            elif default_server_changed:
                 await self.switch_to_interface(server)
             else:
                 await self.switch_lagging_interface()
@@ -712,15 +729,23 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
 
         i = self.interfaces[server]
         if old_interface != i:
+            if not i.is_connected_and_ready():
+                return
             self.logger.info(f"switching to {server}")
-            assert i.ready.done(), "interface we are switching to is not ready yet"
             blockchain_updated = i.blockchain != self.blockchain()
             self.interface = i
-            await i.taskgroup.spawn(self._request_server_info(i))
+            try:
+                await i.taskgroup.spawn(self._request_server_info(i))
+            except RuntimeError as e:  # see #7677
+                if len(e.args) >= 1 and e.args[0] == 'task group terminated':
+                    self.logger.warning(f"tried to switch to {server} but interface.taskgroup is already dead.")
+                    self.interface = None
+                    return
+                raise
             util.trigger_callback('default_server_changed')
             self.default_server_changed_event.set()
             self.default_server_changed_event.clear()
-            self._set_status('connected')
+            self._set_status(ConnectionState.CONNECTED)
             util.trigger_callback('network_updated')
             if blockchain_updated:
                 util.trigger_callback('blockchain_updated')
@@ -758,11 +783,13 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
         We distinguish by whether it is in self.interfaces.'''
         if not interface: return
         if interface.server == self.default_server:
-            self._set_status('disconnected')
+            self._set_status(ConnectionState.DISCONNECTED)
         await self._close_interface(interface)
         util.trigger_callback('network_updated')
 
     def get_network_timeout_seconds(self, request_type=NetworkTimeout.Generic) -> int:
+        if self.config.get('network_timeout', None):
+            return int(self.config.get('network_timeout'))
         if self.oneserver and not self.auto_connect:
             return request_type.MOST_RELAXED
         if self.proxy:
@@ -779,7 +806,7 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
         self._connecting_ifaces.add(server)
         if server == self.default_server:
             self.logger.info(f"connecting to {server} as new interface")
-            self._set_status('connecting')
+            self._set_status(ConnectionState.CONNECTING)
         self._trying_addr_now(server)
 
         interface = Interface(network=self, server=server, proxy=self.proxy)
@@ -844,7 +871,7 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
                 assert iface.ready.done(), "interface not ready yet"
                 # try actual request
                 try:
-                    async with TaskGroup(wait=any) as group:
+                    async with OldTaskGroup(wait=any) as group:
                         task = await group.spawn(func(self, *args, **kwargs))
                         await group.spawn(iface.got_disconnected.wait())
                 except RequestTimedOut:
@@ -860,7 +887,7 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
                 if task.done() and not task.cancelled():
                     return task.result()
                 # otherwise; try again
-            raise BestEffortRequestFailed('no interface to do request on... gave up.')
+            raise BestEffortRequestFailed('cannot establish a connection... gave up.')
         return make_reliable_wrapper
 
     def catch_server_exceptions(func):
@@ -875,10 +902,14 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
     @best_effort_reliable
     @catch_server_exceptions
     async def get_merkle_for_transaction(self, tx_hash: str, tx_height: int) -> dict:
+        if self.interface is None:  # handled by best_effort_reliable
+            raise RequestTimedOut()
         return await self.interface.get_merkle_for_transaction(tx_hash=tx_hash, tx_height=tx_height)
 
     @best_effort_reliable
     async def broadcast_transaction(self, tx: 'Transaction', *, timeout=None) -> None:
+        if self.interface is None:  # handled by best_effort_reliable
+            raise RequestTimedOut()
         if timeout is None:
             timeout = self.get_network_timeout_seconds(NetworkTimeout.Urgent)
         try:
@@ -887,23 +918,26 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
         except (RequestTimedOut, asyncio.CancelledError, asyncio.TimeoutError):
             raise  # pass-through
         except aiorpcx.jsonrpc.CodeMessageError as e:
-            self.logger.info(f"broadcast_transaction error [DO NOT TRUST THIS MESSAGE]: {repr(e)}")
+            self.logger.info(f"broadcast_transaction error [DO NOT TRUST THIS MESSAGE]: {error_text_str_to_safe_str(repr(e))}")
             raise TxBroadcastServerReturnedError(self.sanitize_tx_broadcast_response(e.message)) from e
         except BaseException as e:  # intentional BaseException for sanity!
-            self.logger.info(f"broadcast_transaction error2 [DO NOT TRUST THIS MESSAGE]: {repr(e)}")
+            self.logger.info(f"broadcast_transaction error2 [DO NOT TRUST THIS MESSAGE]: {error_text_str_to_safe_str(repr(e))}")
             send_exception_to_crash_reporter(e)
             raise TxBroadcastUnknownError() from e
         if out != tx.txid():
-            self.logger.info(f"unexpected txid for broadcast_transaction [DO NOT TRUST THIS MESSAGE]: {out} != {tx.txid()}")
+            self.logger.info(f"unexpected txid for broadcast_transaction [DO NOT TRUST THIS MESSAGE]: "
+                             f"{error_text_str_to_safe_str(out)} != {tx.txid()}")
             raise TxBroadcastHashMismatch(_("Server returned unexpected transaction ID."))
 
-    async def try_broadcasting(self, tx, name):
+    async def try_broadcasting(self, tx, name) -> bool:
         try:
             await self.broadcast_transaction(tx)
         except Exception as e:
             self.logger.info(f'error: could not broadcast {name} {tx.txid()}, {str(e)}')
+            return False
         else:
             self.logger.info(f'success: broadcasting {name} {tx.txid()}')
+            return True
 
     @staticmethod
     def sanitize_tx_broadcast_response(server_msg) -> str:
@@ -912,27 +946,7 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
         # server_msg is untrusted input so it should not be shown to the user. see #4968
         server_msg = str(server_msg)
         server_msg = server_msg.replace("\n", r"\n")
-        # https://github.com/bitcoin/bitcoin/blob/5bb64acd9d3ced6e6f95df282a1a0f8b98522cb0/src/policy/policy.cpp
-        # grep "reason ="
-        policy_error_messages = {
-            r"version": _("Transaction uses non-standard version."),
-            r"tx-size": _("The transaction was rejected because it is too large (in bytes)."),
-            r"scriptsig-size": None,
-            r"scriptsig-not-pushonly": None,
-            r"scriptpubkey":
-                ("scriptpubkey\n" +
-                 _("Some of the outputs pay to a non-standard script.")),
-            r"bare-multisig": None,
-            r"dust":
-                (_("Transaction could not be broadcast due to dust outputs.\n"
-                   "Some of the outputs are too small in value, probably lower than 1000 satoshis.\n"
-                   "Check the units, make sure you haven't confused e.g. mBTC and BTC.")),
-            r"multi-op-return": _("The transaction was rejected because it contains multiple OP_RETURN outputs."),
-        }
-        for substring in policy_error_messages:
-            if substring in server_msg:
-                msg = policy_error_messages[substring]
-                return msg if msg else substring
+
         # https://github.com/bitcoin/bitcoin/blob/5bb64acd9d3ced6e6f95df282a1a0f8b98522cb0/src/script/script_error.cpp
         script_error_messages = {
             r"Script evaluated without error but finished with a false/empty top stack element",
@@ -994,7 +1008,7 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
         # https://github.com/bitcoin/bitcoin/blob/5bb64acd9d3ced6e6f95df282a1a0f8b98522cb0/src/validation.cpp
         # grep "REJECT_"
         # grep "TxValidationResult"
-        # should come after script_error.cpp (due to e.g. non-mandatory-script-verify-flag)
+        # should come after script_error.cpp (due to e.g. "non-mandatory-script-verify-flag")
         validation_error_messages = {
             r"coinbase": None,
             r"tx-size-small": None,
@@ -1076,37 +1090,72 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
             if substring in server_msg:
                 msg = tx_verify_error_messages[substring]
                 return msg if msg else substring
+        # https://github.com/bitcoin/bitcoin/blob/5bb64acd9d3ced6e6f95df282a1a0f8b98522cb0/src/policy/policy.cpp
+        # grep "reason ="
+        # should come after validation.cpp (due to "tx-size" vs "tx-size-small")
+        # should come after script_error.cpp (due to e.g. "version")
+        policy_error_messages = {
+            r"version": _("Transaction uses non-standard version."),
+            r"tx-size": _("The transaction was rejected because it is too large (in bytes)."),
+            r"scriptsig-size": None,
+            r"scriptsig-not-pushonly": None,
+            r"scriptpubkey":
+                ("scriptpubkey\n" +
+                 _("Some of the outputs pay to a non-standard script.")),
+            r"bare-multisig": None,
+            r"dust":
+                (_("Transaction could not be broadcast due to dust outputs.\n"
+                   "Some of the outputs are too small in value, probably lower than 1000 satoshis.\n"
+                   "Check the units, make sure you haven't confused e.g. mBTC and BTC.")),
+            r"multi-op-return": _("The transaction was rejected because it contains multiple OP_RETURN outputs."),
+        }
+        for substring in policy_error_messages:
+            if substring in server_msg:
+                msg = policy_error_messages[substring]
+                return msg if msg else substring
         # otherwise:
         return _("Unknown error")
 
     @best_effort_reliable
     @catch_server_exceptions
     async def request_chunk(self, height: int, tip=None, *, can_return_early=False):
+        if self.interface is None:  # handled by best_effort_reliable
+            raise RequestTimedOut()
         return await self.interface.request_chunk(height, tip=tip, can_return_early=can_return_early)
 
     @best_effort_reliable
     @catch_server_exceptions
     async def get_transaction(self, tx_hash: str, *, timeout=None) -> str:
+        if self.interface is None:  # handled by best_effort_reliable
+            raise RequestTimedOut()
         return await self.interface.get_transaction(tx_hash=tx_hash, timeout=timeout)
 
     @best_effort_reliable
     @catch_server_exceptions
     async def get_history_for_scripthash(self, sh: str) -> List[dict]:
+        if self.interface is None:  # handled by best_effort_reliable
+            raise RequestTimedOut()
         return await self.interface.get_history_for_scripthash(sh)
 
     @best_effort_reliable
     @catch_server_exceptions
     async def listunspent_for_scripthash(self, sh: str) -> List[dict]:
+        if self.interface is None:  # handled by best_effort_reliable
+            raise RequestTimedOut()
         return await self.interface.listunspent_for_scripthash(sh)
 
     @best_effort_reliable
     @catch_server_exceptions
     async def get_balance_for_scripthash(self, sh: str) -> dict:
+        if self.interface is None:  # handled by best_effort_reliable
+            raise RequestTimedOut()
         return await self.interface.get_balance_for_scripthash(sh)
 
     @best_effort_reliable
     @catch_server_exceptions
     async def get_txid_from_txpos(self, tx_height, tx_pos, merkle):
+        if self.interface is None:  # handled by best_effort_reliable
+            raise RequestTimedOut()
         return await self.interface.get_txid_from_txpos(tx_height, tx_pos, merkle)
 
     def blockchain(self) -> Blockchain:
@@ -1186,30 +1235,27 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
 
     async def _start(self):
         assert not self.taskgroup
-        self.taskgroup = taskgroup = SilentTaskGroup()
+        self.taskgroup = taskgroup = OldTaskGroup()
         assert not self.interface and not self.interfaces
         assert not self._connecting_ifaces
         assert not self._closing_ifaces
         self.logger.info('starting network')
         self._clear_addr_retry_times()
-        self._set_proxy(deserialize_proxy(self.config.get('proxy')))
-        self._maybe_set_oneserver()
+        self._init_parameters_from_config()
         await self.taskgroup.spawn(self._run_new_interface(self.default_server))
 
         async def main():
-            self.logger.info("starting taskgroup.")
+            self.logger.info(f"starting taskgroup ({hex(id(taskgroup))}).")
             try:
                 # note: if a task finishes with CancelledError, that
                 # will NOT raise, and the group will keep the other tasks running
                 async with taskgroup as group:
                     await group.spawn(self._maintain_sessions())
                     [await group.spawn(job) for job in self._jobs]
-            except asyncio.CancelledError:
-                raise
             except Exception as e:
-                self.logger.exception("taskgroup died.")
+                self.logger.exception(f"taskgroup died ({hex(id(taskgroup))}).")
             finally:
-                self.logger.info("taskgroup stopped.")
+                self.logger.info(f"taskgroup stopped ({hex(id(taskgroup))}).")
         asyncio.run_coroutine_threadsafe(main(), self.asyncio_loop)
 
         util.trigger_callback('network_updated')
@@ -1220,16 +1266,20 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
         Note: the jobs will *restart* every time the network restarts, e.g. on proxy
         setting changes.
         """
+        self._was_started = True
         self._jobs = jobs or []
         asyncio.run_coroutine_threadsafe(self._start(), self.asyncio_loop)
 
     @log_exceptions
     async def stop(self, *, full_shutdown: bool = True):
+        if not self._was_started:
+            self.logger.info("not stopping network as it was never started")
+            return
         self.logger.info("stopping network")
         # timeout: if full_shutdown, it is up to the caller to time us out,
         #          otherwise if e.g. restarting due to proxy changes, we time out fast
         async with (nullcontext() if full_shutdown else ignore_after(1)):
-            async with TaskGroup() as group:
+            async with OldTaskGroup() as group:
                 await group.spawn(self.taskgroup.cancel_remaining())
                 if full_shutdown:
                     await group.spawn(self.stop_gossip(full_shutdown=full_shutdown))
@@ -1242,13 +1292,13 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
             util.trigger_callback('network_updated')
 
     async def _ensure_there_is_a_main_interface(self):
-        if self.is_connected():
+        if self.interface:
             return
         # if auto_connect is set, try a different server
         if self.auto_connect and not self.is_connecting():
             await self._switch_to_random_interface()
         # if auto_connect is not set, or still no main interface, retry current
-        if not self.is_connected() and not self.is_connecting():
+        if not self.interface and not self.is_connecting():
             if self._can_retry_addr(self.default_server, urgent=True):
                 await self.switch_to_interface(self.default_server)
 
@@ -1275,21 +1325,21 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
                     await self.interface.taskgroup.spawn(self._request_fee_estimates, self.interface)
 
         while True:
-            try:
-                await maybe_start_new_interfaces()
-                await maintain_healthy_spread_of_connected_servers()
-                await maintain_main_interface()
-            except asyncio.CancelledError:
-                # suppress spurious cancellations
-                group = self.taskgroup
-                if not group or group.closed():
-                    raise
+            await maybe_start_new_interfaces()
+            await maintain_healthy_spread_of_connected_servers()
+            await maintain_main_interface()
             await asyncio.sleep(0.1)
 
     @classmethod
-    async def _send_http_on_proxy(cls, method: str, url: str, params: str = None,
-                                  body: bytes = None, json: dict = None, headers=None,
-                                  on_finish=None, timeout=None):
+    async def async_send_http_on_proxy(
+            cls, method: str, url: str, *,
+            params: dict = None,
+            body: bytes = None,
+            json: dict = None,
+            headers=None,
+            on_finish=None,
+            timeout=None,
+    ):
         async def default_on_finish(resp: ClientResponse):
             resp.raise_for_status()
             return await resp.text()
@@ -1318,11 +1368,11 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
     def send_http_on_proxy(cls, method, url, **kwargs):
         network = cls.get_instance()
         if network:
-            assert network._loop_thread is not threading.currentThread()
+            assert util.get_running_loop() != network.asyncio_loop
             loop = network.asyncio_loop
         else:
-            loop = asyncio.get_event_loop()
-        coro = asyncio.run_coroutine_threadsafe(cls._send_http_on_proxy(method, url, **kwargs), loop)
+            loop = util.get_asyncio_loop()
+        coro = asyncio.run_coroutine_threadsafe(cls.async_send_http_on_proxy(method, url, **kwargs), loop)
         # note: _send_http_on_proxy has its own timeout, so no timeout here:
         return coro.result()
 
@@ -1333,11 +1383,19 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
         session = self.interface.session
         return parse_servers(await session.send_request('server.peers.subscribe'))
 
-    async def send_multiple_requests(self, servers: Sequence[ServerAddr], method: str, params: Sequence):
+    async def send_multiple_requests(
+            self,
+            servers: Sequence[ServerAddr],
+            method: str,
+            params: Sequence,
+            *,
+            timeout: int = None,
+    ):
+        if timeout is None:
+            timeout = self.get_network_timeout_seconds(NetworkTimeout.Urgent)
         responses = dict()
         async def get_response(server: ServerAddr):
             interface = Interface(network=self, server=server, proxy=self.proxy)
-            timeout = self.get_network_timeout_seconds(NetworkTimeout.Urgent)
             try:
                 await asyncio.wait_for(interface.ready, timeout)
             except BaseException as e:
@@ -1348,7 +1406,16 @@ class Network(Logger, NetworkRetryManager[ServerAddr]):
             except Exception as e:
                 res = e
             responses[interface.server] = res
-        async with TaskGroup() as group:
+        async with OldTaskGroup() as group:
             for server in servers:
                 await group.spawn(get_response(server))
         return responses
+
+    async def prune_offline_servers(self, hostmap):
+        peers = filter_protocol(hostmap, allowed_protocols=("t", "s",))
+        timeout = self.get_network_timeout_seconds(NetworkTimeout.Generic)
+        replies = await self.send_multiple_requests(peers, 'blockchain.headers.subscribe', [], timeout=timeout)
+        servers_replied = {serveraddr.host for serveraddr in replies.keys()}
+        servers_dict = {k: v for k, v in hostmap.items()
+                        if k in servers_replied}
+        return servers_dict
