@@ -10,11 +10,11 @@ import struct
 from electrum import bip32
 from electrum.bip32 import BIP32Node, InvalidMasterKeyVersionBytes
 from electrum.i18n import _
-from electrum.plugin import Device, hook
+from electrum.plugin import Device, hook, runs_in_hwd_thread
 from electrum.keystore import Hardware_KeyStore, KeyStoreWithMPK
 from electrum.transaction import PartialTransaction
 from electrum.wallet import Standard_Wallet, Multisig_Wallet, Abstract_Wallet
-from electrum.util import bfh, bh2u, versiontuple, UserFacingException
+from electrum.util import bfh, versiontuple, UserFacingException
 from electrum.base_wizard import ScriptTypeNotSupported
 from electrum.logging import get_logger
 
@@ -43,13 +43,11 @@ try:
             # verify a signature (65 bytes) over the session key, using the master bip32 node
             # - customized to use specific EC library of Electrum.
             pubkey = BIP32Node.from_xkey(expect_xpub).eckey
-            try:
-                pubkey.verify_message_hash(sig[1:65], self.session_key)
-                return True
-            except:
-                return False
+            return pubkey.verify_message_hash(sig[1:65], self.session_key)
 
-except ImportError:
+except ImportError as e:
+    if not (isinstance(e, ModuleNotFoundError) and e.name == 'ckcc'):
+        _logger.exception('error importing coldcard plugin deps')
     requirements_ok = False
 
     COINKITE_VID = 0xd13e
@@ -72,9 +70,8 @@ class CKCCClient(HardwareClientBase):
             self.dev = ElectrumColdcardDevice(dev_path, encrypt=True)
         else:
             # open the real HID device
-            with self.device_manager().hid_lock:
-                hd = hid.device(path=dev_path)
-                hd.open_path(dev_path)
+            hd = hid.device(path=dev_path)
+            hd.open_path(dev_path)
 
             self.dev = ElectrumColdcardDevice(dev=hd, encrypt=True)
 
@@ -85,6 +82,7 @@ class CKCCClient(HardwareClientBase):
         return '<CKCCClient: xfp=%s label=%r>' % (xfp2str(self.dev.master_fingerprint),
                                                         self.label())
 
+    @runs_in_hwd_thread
     def verify_connection(self, expected_xfp: int, expected_xpub=None):
         ex = (expected_xfp, expected_xpub)
 
@@ -95,7 +93,7 @@ class CKCCClient(HardwareClientBase):
         if expected_xpub is None:
             expected_xpub = self.dev.master_xpub
 
-        if ( (self._expected_device is not None) 
+        if ((self._expected_device is not None)
                 or (self.dev.master_fingerprint != expected_xfp)
                 or (self.dev.master_xpub != expected_xpub)):
             # probably indicating programing error, not hacking
@@ -121,14 +119,10 @@ class CKCCClient(HardwareClientBase):
         # can't do anything w/ devices that aren't setup (this code not normally reachable)
         return bool(self.dev.master_xpub)
 
-    def timeout(self, cutoff):
-        # nothing to do?
-        pass
-
+    @runs_in_hwd_thread
     def close(self):
         # close the HID device (so can be reused)
-        with self.device_manager().hid_lock:
-            self.dev.close()
+        self.dev.close()
         self.dev = None
 
     def is_initialized(self):
@@ -146,20 +140,21 @@ class CKCCClient(HardwareClientBase):
         else:
             lab = 'Coldcard ' + xfp2str(self.dev.master_fingerprint)
 
-        # Hack zone: during initial setup I need the xfp and master xpub but 
-        # very few objects are passed between the various steps of base_wizard.
-        # Solution: return a string with some hidden metadata
-        # - see <https://stackoverflow.com/questions/7172772/abc-for-string>
-        # - needs to work w/ deepcopy
-        class LabelStr(str):
-            def __new__(cls, s, xfp=None, xpub=None):
-                self = super().__new__(cls, str(s))
-                self.xfp = getattr(s, 'xfp', xfp)
-                self.xpub = getattr(s, 'xpub', xpub)
-                return self
+        return lab
 
-        return LabelStr(lab, self.dev.master_fingerprint, self.dev.master_xpub)
+    def manipulate_keystore_dict_during_wizard_setup(self, d: dict):
+        master_xpub = self.dev.master_xpub
+        if master_xpub is not None:
+            try:
+                node = BIP32Node.from_xkey(master_xpub)
+            except InvalidMasterKeyVersionBytes:
+                raise UserFacingException(
+                    _('Invalid xpub magic. Make sure your {} device is set to the correct chain.').format(self.device) + ' ' +
+                    _('You might have to unplug and plug it in again.')
+                ) from None
+            d['ckcc_xpub'] = master_xpub
 
+    @runs_in_hwd_thread
     def has_usable_connection_with_device(self):
         # Do end-to-end ping test
         try:
@@ -168,6 +163,7 @@ class CKCCClient(HardwareClientBase):
         except:
             return False
 
+    @runs_in_hwd_thread
     def get_xpub(self, bip32_path, xtype):
         assert xtype in ColdcardPlugin.SUPPORTED_XTYPES
         _logger.info('Derive xtype = %r' % xtype)
@@ -183,6 +179,7 @@ class CKCCClient(HardwareClientBase):
             xpub = node._replace(xtype=xtype).to_xpub()
         return xpub
 
+    @runs_in_hwd_thread
     def ping_check(self):
         # check connection is working
         assert self.dev.session_key, 'not encrypted?'
@@ -193,26 +190,32 @@ class CKCCClient(HardwareClientBase):
         except:
             raise RuntimeError("Communication trouble with Coldcard")
 
+    @runs_in_hwd_thread
     def show_address(self, path, addr_fmt):
         # prompt user w/ address, also returns it immediately.
         return self.dev.send_recv(CCProtocolPacker.show_address(path, addr_fmt), timeout=None)
 
+    @runs_in_hwd_thread
     def show_p2sh_address(self, *args, **kws):
         # prompt user w/ p2sh address, also returns it immediately.
         return self.dev.send_recv(CCProtocolPacker.show_p2sh_address(*args, **kws), timeout=None)
 
+    @runs_in_hwd_thread
     def get_version(self):
         # gives list of strings
         return self.dev.send_recv(CCProtocolPacker.version(), timeout=1000).split('\n')
 
+    @runs_in_hwd_thread
     def sign_message_start(self, path, msg):
         # this starts the UX experience.
         self.dev.send_recv(CCProtocolPacker.sign_message(msg, path), timeout=None)
 
+    @runs_in_hwd_thread
     def sign_message_poll(self):
         # poll device... if user has approved, will get tuple: (addr, sig) else None
         return self.dev.send_recv(CCProtocolPacker.get_signed_msg(), timeout=None)
 
+    @runs_in_hwd_thread
     def sign_transaction_start(self, raw_psbt: bytes, *, finalize: bool = False):
         # Multiple steps to sign:
         # - upload binary
@@ -225,18 +228,20 @@ class CKCCClient(HardwareClientBase):
         resp = self.dev.send_recv(CCProtocolPacker.sign_transaction(dlen, chk, finalize=finalize),
                                     timeout=None)
 
-        if resp != None:
+        if resp is not None:
             raise ValueError(resp)
 
+    @runs_in_hwd_thread
     def sign_transaction_poll(self):
         # poll device... if user has approved, will get tuple: (legnth, checksum) else None
         return self.dev.send_recv(CCProtocolPacker.get_signed_txn(), timeout=None)
 
+    @runs_in_hwd_thread
     def download_file(self, length, checksum, file_number=1):
         # get a file
         return self.dev.download_file(length, checksum, file_number=file_number)
 
-        
+
 
 class Coldcard_KeyStore(Hardware_KeyStore):
     hw_type = 'coldcard'
@@ -246,17 +251,12 @@ class Coldcard_KeyStore(Hardware_KeyStore):
 
     def __init__(self, d):
         Hardware_KeyStore.__init__(self, d)
-        # Errors and other user interaction is done through the wallet's
-        # handler.  The handler is per-window and preserved across
-        # device reconnects
-        self.force_watching_only = False
         self.ux_busy = False
 
         # we need to know at least the fingerprint of the master xpub to verify against MiTM
         # - device reports these value during encryption setup process
         # - full xpub value now optional
-        lab = d['label']
-        self.ckcc_xpub = getattr(lab, 'xpub', None) or d.get('ckcc_xpub', None)
+        self.ckcc_xpub = d.get('ckcc_xpub', None)
 
     def dump(self):
         # our additions to the stored data about keystore -- only during creation?
@@ -269,25 +269,23 @@ class Coldcard_KeyStore(Hardware_KeyStore):
         assert xfp is not None
         return xfp_int_from_xfp_bytes(bfh(xfp))
 
-    def get_client(self):
+    def get_client(self, *args, **kwargs):
         # called when user tries to do something like view address, sign somthing.
         # - not called during probing/setup
         # - will fail if indicated device can't produce the xpub (at derivation) expected
-        rv = self.plugin.get_client(self)
-        if rv:
+        client = super().get_client(*args, **kwargs)  # type: Optional[CKCCClient]
+        if client:
             xfp_int = self.get_xfp_int()
-            rv.verify_connection(xfp_int, self.ckcc_xpub)
+            client.verify_connection(xfp_int, self.ckcc_xpub)
 
-        return rv
+        return client
 
-    def give_error(self, message, clear_client=False):
+    def give_error(self, message):
         self.logger.info(message)
         if not self.ux_busy:
             self.handler.show_error(message)
         else:
             self.ux_busy = False
-        if clear_client:
-            self.client = None
         raise UserFacingException(message)
 
     def wrap_busy(func):
@@ -304,7 +302,7 @@ class Coldcard_KeyStore(Hardware_KeyStore):
         raise UserFacingException(_('Encryption and decryption are currently not supported for {}').format(self.device))
 
     @wrap_busy
-    def sign_message(self, sequence, message, password):
+    def sign_message(self, sequence, message, password, *, script_type=None):
         # Sign a message on device. Since we have big screen, of course we
         # have to show the message unabiguously there first!
         try:
@@ -313,11 +311,10 @@ class Coldcard_KeyStore(Hardware_KeyStore):
         except (UnicodeError, AssertionError):
             # there are other restrictions on message content,
             # but let the device enforce and report those
-            self.handler.show_error('Only short (%d max) ASCII messages can be signed.' 
+            self.handler.show_error('Only short (%d max) ASCII messages can be signed.'
                                             % MSG_SIGNING_MAX_LENGTH)
             return b''
 
-        client = self.get_client()
         path = self.get_derivation_prefix() + ("/%d/%d" % sequence)
         try:
             cl = self.get_client()
@@ -352,7 +349,7 @@ class Coldcard_KeyStore(Hardware_KeyStore):
             self.handler.show_error('{}\n\n{}'.format(
                 _('Error showing address') + ':', str(exc)))
         except Exception as e:
-            self.give_error(e, True)
+            self.give_error(e)
 
         # give empty bytes for error cases; it seems to clear the old signature box
         return b''
@@ -385,7 +382,7 @@ class Coldcard_KeyStore(Hardware_KeyStore):
                         break
 
                 rlen, rsha = resp
-            
+
                 # download the resulting txn.
                 raw_resp = client.download_file(rlen, rsha)
 
@@ -398,7 +395,7 @@ class Coldcard_KeyStore(Hardware_KeyStore):
             return
         except BaseException as e:
             self.logger.exception('')
-            self.give_error(e, True)
+            self.give_error(e)
             return
 
         tx2 = PartialTransaction.from_raw_psbt(raw_resp)
@@ -508,10 +505,8 @@ class ColdcardPlugin(HW_PluginBase):
 
         return []
 
+    @runs_in_hwd_thread
     def create_client(self, device, handler):
-        if handler:
-            self.handler = handler
-
         # We are given a HID device, or at least some details about it.
         # Not sure why not we aren't just given a HID library handle, but
         # the 'path' is unabiguous, so we'll use that.
@@ -539,6 +534,7 @@ class ColdcardPlugin(HW_PluginBase):
         xpub = client.get_xpub(derivation, xtype)
         return xpub
 
+    @runs_in_hwd_thread
     def get_client(self, keystore, force_pair=True, *,
                    devices=None, allow_user_interaction=True) -> Optional['CKCCClient']:
         # Acquire a connection to the hardware device (via USB)
@@ -560,31 +556,24 @@ class ColdcardPlugin(HW_PluginBase):
         print('# Exported from Electrum', file=fp)
         print(f'Name: {name:.20s}', file=fp)
         print(f'Policy: {wallet.m} of {wallet.n}', file=fp)
-        print(f'Format: {wallet.txin_type.upper()}' , file=fp)
+        print(f'Format: {wallet.txin_type.upper()}', file=fp)
 
         xpubs = []
-        derivs = set()
         for xpub, ks in zip(wallet.get_master_public_keys(), wallet.get_keystores()):  # type: str, KeyStoreWithMPK
             fp_bytes, der_full = ks.get_fp_and_derivation_to_be_used_in_partial_tx(der_suffix=[], only_der_suffix=False)
             fp_hex = fp_bytes.hex().upper()
             der_prefix_str = bip32.convert_bip32_intpath_to_strpath(der_full)
-            xpubs.append( (fp_hex, xpub, der_prefix_str) )
-            derivs.add(der_prefix_str)
+            xpubs.append((fp_hex, xpub, der_prefix_str))
 
-        # Derivation doesn't matter too much to the Coldcard, since it
-        # uses key path data from PSBT or USB request as needed. However,
-        # if there is a clear value, provide it.
-        if len(derivs) == 1:
-            print("Derivation: " + derivs.pop(), file=fp)
+        # Before v3.2.1 derivation didn't matter too much to the Coldcard, since it
+        # could use key path data from PSBT or USB request as needed. However,
+        # derivation data is now required.
 
         print('', file=fp)
 
         assert len(xpubs) == wallet.n
         for xfp, xpub, der_prefix in xpubs:
-            if derivs:
-                # show as a comment if unclear
-                print(f'# derivation: {der_prefix}', file=fp)
-
+            print(f'Derivation: {der_prefix}', file=fp)
             print(f'{xfp}: {xpub}\n', file=fp)
 
     def show_address(self, wallet, address, keystore: 'Coldcard_KeyStore' = None):
@@ -608,7 +597,8 @@ class ColdcardPlugin(HW_PluginBase):
             pubkey_deriv_info = wallet.get_public_keys_with_deriv_info(address)
             pubkey_hexes = sorted([pk.hex() for pk in list(pubkey_deriv_info)])
             xfp_paths = []
-            for pubkey in pubkey_deriv_info:
+            for pubkey_hex in pubkey_hexes:
+                pubkey = bytes.fromhex(pubkey_hex)
                 ks, der_suffix = pubkey_deriv_info[pubkey]
                 fp_bytes, der_full = ks.get_fp_and_derivation_to_be_used_in_partial_tx(der_suffix, only_der_suffix=False)
                 xfp_int = xfp_int_from_xfp_bytes(fp_bytes)
