@@ -24,25 +24,30 @@
 # SOFTWARE.
 import os
 import ast
+import datetime
 import json
 import copy
 import threading
 from collections import defaultdict
 from typing import Dict, Optional, List, Tuple, Set, Iterable, NamedTuple, Sequence, TYPE_CHECKING, Union
 import binascii
+import time
+
+import attr
 
 from . import util, bitcoin
 from .util import profiler, WalletFileException, multisig_type, TxMinedInfo, bfh
-from .invoices import PR_TYPE_ONCHAIN, Invoice
+from .invoices import Invoice, Request
 from .keystore import bip44_derivation
 from .transaction import Transaction, TxOutpoint, tx_from_any, PartialTransaction, PartialTxOutput
 from .logging import Logger
-from .lnutil import LOCAL, REMOTE, FeeUpdate, UpdateAddHtlc, LocalConfig, RemoteConfig, Keypair, OnlyPubkeyKeypair, RevocationStore, ChannelBackupStorage
+from .lnutil import LOCAL, REMOTE, FeeUpdate, UpdateAddHtlc, LocalConfig, RemoteConfig, ChannelType
+from .lnutil import ImportedChannelBackupStorage, OnchainChannelBackupStorage
 from .lnutil import ChannelConstraints, Outpoint, ShachainElement
-from .json_db import StoredDict, JsonDB, locked, modifier
+from .json_db import StoredDict, JsonDB, locked, modifier, StoredObject
 from .plugin import run_hook, plugin_loaders
-from .paymentrequest import PaymentRequest
 from .submarine_swaps import SwapData
+from .version import ELECTRUM_VERSION
 
 if TYPE_CHECKING:
     from .storage import WalletStorage
@@ -52,7 +57,7 @@ if TYPE_CHECKING:
 
 OLD_SEED_VERSION = 4        # electrum versions < 2.0
 NEW_SEED_VERSION = 11       # electrum versions >= 2.0
-FINAL_SEED_VERSION = 33     # electrum >= 2.7 will set this to prevent
+FINAL_SEED_VERSION = 51     # electrum >= 2.7 will set this to prevent
                             # old versions from overwriting new format
 
 
@@ -62,6 +67,20 @@ class TxFeesValue(NamedTuple):
     num_inputs: Optional[int] = None
 
 
+@attr.s
+class DBMetadata(StoredObject):
+    creation_timestamp = attr.ib(default=None, type=int)
+    first_electrum_version_used = attr.ib(default=None, type=str)
+
+    def to_str(self) -> str:
+        ts = self.creation_timestamp
+        ver = self.first_electrum_version_used
+        if ts is None or ver is None:
+            return "unknown"
+        date_str = datetime.date.fromtimestamp(ts).isoformat()
+        return f"using {ver}, on {date_str}"
+
+
 class WalletDB(JsonDB):
 
     def __init__(self, raw, *, manual_upgrades: bool):
@@ -73,6 +92,7 @@ class WalletDB(JsonDB):
             self.load_plugins()
         else:  # creating new db
             self.put('seed_version', FINAL_SEED_VERSION)
+            self._add_db_creation_metadata()
             self._after_upgrade_tasks()
 
     def load_data(self, s):
@@ -129,7 +149,8 @@ class WalletDB(JsonDB):
             data2['suffix'] = 'imported'
             result = [data1, data2]
 
-        elif wallet_type in ['bip44', 'trezor', 'keepkey', 'ledger', 'btchip', 'digitalbitbox', 'safe_t']:
+        # note: do not add new hardware types here, this code is for converting legacy wallets
+        elif wallet_type in ['bip44', 'trezor', 'keepkey', 'ledger', 'btchip']:
             mpk = self.get('master_public_keys')
             for k in d.keys():
                 i = int(k)
@@ -181,6 +202,24 @@ class WalletDB(JsonDB):
         self._convert_version_31()
         self._convert_version_32()
         self._convert_version_33()
+        self._convert_version_34()
+        self._convert_version_35()
+        self._convert_version_36()
+        self._convert_version_37()
+        self._convert_version_38()
+        self._convert_version_39()
+        self._convert_version_40()
+        self._convert_version_41()
+        self._convert_version_42()
+        self._convert_version_43()
+        self._convert_version_44()
+        self._convert_version_45()
+        self._convert_version_46()
+        self._convert_version_47()
+        self._convert_version_48()
+        self._convert_version_49()
+        self._convert_version_50()
+        self._convert_version_51()
         self.put('seed_version', FINAL_SEED_VERSION)  # just to be sure
 
         self._after_upgrade_tasks()
@@ -245,7 +284,8 @@ class WalletDB(JsonDB):
             self.put('wallet_type', 'standard')
             self.put('keystore', d)
 
-        elif wallet_type in ['trezor', 'keepkey', 'ledger', 'digitalbitbox', 'safe_t']:
+        # note: do not add new hardware types here, this code is for converting legacy wallets
+        elif wallet_type in ['trezor', 'keepkey', 'ledger']:
             xpub = xpubs["x/0'"]
             derivation = self.get('derivation', bip44_derivation(0))
             d = {
@@ -521,10 +561,10 @@ class WalletDB(JsonDB):
             r['buckets'] = d
             c['revocation_store'] = r
         # convert channels to dict
-        self.data['channels'] = { x['channel_id']: x for x in channels }
+        self.data['channels'] = {x['channel_id']: x for x in channels}
         # convert txi & txo
         txi = self.get('txi', {})
-        for tx_hash, d in txi.items():
+        for tx_hash, d in list(txi.items()):
             d2 = {}
             for addr, l in d.items():
                 d2[addr] = {}
@@ -533,7 +573,7 @@ class WalletDB(JsonDB):
             txi[tx_hash] = d2
         self.data['txi'] = txi
         txo = self.get('txo', {})
-        for tx_hash, d in txo.items():
+        for tx_hash, d in list(txo.items()):
             d2 = {}
             for addr, l in d.items():
                 d2[addr] = {}
@@ -545,9 +585,11 @@ class WalletDB(JsonDB):
         self.data['seed_version'] = 24
 
     def _convert_version_25(self):
+        from .crypto import sha256
         if not self._is_upgrade_method_needed(24, 24):
             return
         # add 'type' field to onchain requests
+        PR_TYPE_ONCHAIN = 0
         requests = self.data.get('payment_requests', {})
         for k, r in list(requests.items()):
             if r.get('address') == k:
@@ -560,25 +602,15 @@ class WalletDB(JsonDB):
                     'time': r.get('time'),
                     'type': PR_TYPE_ONCHAIN,
                 }
-        # convert bip70 invoices
+        # delete bip70 invoices
+        # note: this upgrade was changed ~2 years after-the-fact to delete instead of converting
         invoices = self.data.get('invoices', {})
         for k, r in list(invoices.items()):
             data = r.get("hex")
-            if data:
-                pr = PaymentRequest(bytes.fromhex(data))
-                if pr.id != k:
-                    continue
-                invoices[k] = {
-                    'type': PR_TYPE_ONCHAIN,
-                    'amount': pr.get_amount(),
-                    'bip70': data,
-                    'exp': pr.get_expiration_date() - pr.get_time(),
-                    'id': pr.id,
-                    'message': pr.get_memo(),
-                    'outputs': [x.to_legacy_tuple() for x in pr.get_outputs()],
-                    'time': pr.get_time(),
-                    'requestor': pr.get_requestor(),
-                }
+            pr_id = sha256(bytes.fromhex(data))[0:16].hex()
+            if pr_id != k:
+                continue
+            del invoices[k]
         self.data['seed_version'] = 25
 
     def _convert_version_26(self):
@@ -615,6 +647,7 @@ class WalletDB(JsonDB):
     def _convert_version_29(self):
         if not self._is_upgrade_method_needed(28, 28):
             return
+        PR_TYPE_ONCHAIN = 0
         requests = self.data.get('payment_requests', {})
         invoices = self.data.get('invoices', {})
         for d in [invoices, requests]:
@@ -650,8 +683,8 @@ class WalletDB(JsonDB):
     def _convert_version_30(self):
         if not self._is_upgrade_method_needed(29, 29):
             return
-
-        from .invoices import PR_TYPE_ONCHAIN, PR_TYPE_LN
+        PR_TYPE_ONCHAIN = 0
+        PR_TYPE_LN = 2
         requests = self.data.get('payment_requests', {})
         invoices = self.data.get('invoices', {})
         for d in [invoices, requests]:
@@ -673,8 +706,7 @@ class WalletDB(JsonDB):
     def _convert_version_31(self):
         if not self._is_upgrade_method_needed(30, 30):
             return
-
-        from .invoices import PR_TYPE_ONCHAIN
+        PR_TYPE_ONCHAIN = 0
         requests = self.data.get('payment_requests', {})
         invoices = self.data.get('invoices', {})
         for d in [invoices, requests]:
@@ -688,8 +720,7 @@ class WalletDB(JsonDB):
     def _convert_version_32(self):
         if not self._is_upgrade_method_needed(31, 31):
             return
-
-        from .invoices import PR_TYPE_ONCHAIN
+        PR_TYPE_ONCHAIN = 0
         invoices_old = self.data.get('invoices', {})
         invoices_new = {k: item for k, item in invoices_old.items()
                         if not (item['type'] == PR_TYPE_ONCHAIN and item['outputs'] is None)}
@@ -699,17 +730,292 @@ class WalletDB(JsonDB):
     def _convert_version_33(self):
         if not self._is_upgrade_method_needed(32, 32):
             return
-
-        from .invoices import PR_TYPE_ONCHAIN
+        PR_TYPE_ONCHAIN = 0
         requests = self.data.get('payment_requests', {})
         invoices = self.data.get('invoices', {})
         for d in [invoices, requests]:
             for key, item in list(d.items()):
                 if item['type'] == PR_TYPE_ONCHAIN:
                     item['height'] = item.get('height') or 0
-
         self.data['seed_version'] = 33
 
+    def _convert_version_34(self):
+        if not self._is_upgrade_method_needed(33, 33):
+            return
+        channels = self.data.get('channels', {})
+        for key, item in channels.items():
+            item['local_config']['upfront_shutdown_script'] = \
+                item['local_config'].get('upfront_shutdown_script') or ""
+            item['remote_config']['upfront_shutdown_script'] = \
+                item['remote_config'].get('upfront_shutdown_script') or ""
+        self.data['seed_version'] = 34
+
+    def _convert_version_35(self):
+        # same as 32, but for payment_requests
+        if not self._is_upgrade_method_needed(34, 34):
+            return
+        PR_TYPE_ONCHAIN = 0
+        requests_old = self.data.get('payment_requests', {})
+        requests_new = {k: item for k, item in requests_old.items()
+                        if not (item['type'] == PR_TYPE_ONCHAIN and item['outputs'] is None)}
+        self.data['payment_requests'] = requests_new
+        self.data['seed_version'] = 35
+
+    def _convert_version_36(self):
+        if not self._is_upgrade_method_needed(35, 35):
+            return
+        old_frozen_coins = self.data.get('frozen_coins', [])
+        new_frozen_coins = {coin: True for coin in old_frozen_coins}
+        self.data['frozen_coins'] = new_frozen_coins
+        self.data['seed_version'] = 36
+
+    def _convert_version_37(self):
+        if not self._is_upgrade_method_needed(36, 36):
+            return
+        payments = self.data.get('lightning_payments', {})
+        for k, v in list(payments.items()):
+            amount_sat, direction, status = v
+            amount_msat = amount_sat * 1000 if amount_sat is not None else None
+            payments[k] = amount_msat, direction, status
+        self.data['lightning_payments'] = payments
+        self.data['seed_version'] = 37
+
+    def _convert_version_38(self):
+        if not self._is_upgrade_method_needed(37, 37):
+            return
+        PR_TYPE_ONCHAIN = 0
+        PR_TYPE_LN = 2
+        from .bitcoin import TOTAL_COIN_SUPPLY_LIMIT_IN_BTC, COIN
+        max_sats = TOTAL_COIN_SUPPLY_LIMIT_IN_BTC * COIN
+        requests = self.data.get('payment_requests', {})
+        invoices = self.data.get('invoices', {})
+        for d in [invoices, requests]:
+            for key, item in list(d.items()):
+                if item['type'] == PR_TYPE_ONCHAIN:
+                    amount_sat = item['amount_sat']
+                    if amount_sat == '!':
+                        continue
+                    if not (isinstance(amount_sat, int) and 0 <= amount_sat <= max_sats):
+                        del d[key]
+                elif item['type'] == PR_TYPE_LN:
+                    amount_msat = item['amount_msat']
+                    if not amount_msat:
+                        continue
+                    if not (isinstance(amount_msat, int) and 0 <= amount_msat <= max_sats * 1000):
+                        del d[key]
+        self.data['seed_version'] = 38
+
+    def _convert_version_39(self):
+        # this upgrade prevents initialization of lightning_privkey2 after lightning_xprv has been set
+        if not self._is_upgrade_method_needed(38, 38):
+            return
+        self.data['imported_channel_backups'] = self.data.pop('channel_backups', {})
+        self.data['seed_version'] = 39
+
+    def _convert_version_40(self):
+        # put 'seed_type' into keystores
+        if not self._is_upgrade_method_needed(39, 39):
+            return
+        for ks_name in ('keystore', *['x{}/'.format(i) for i in range(1, 16)]):
+            ks = self.data.get(ks_name, None)
+            if ks is None: continue
+            seed = ks.get('seed')
+            if not seed: continue
+            seed_type = None
+            xpub = ks.get('xpub') or None
+            if xpub:
+                assert isinstance(xpub, str)
+                if xpub[0:4] in ('xpub', 'tpub'):
+                    seed_type = 'standard'
+                elif xpub[0:4] in ('zpub', 'Zpub', 'vpub', 'Vpub'):
+                    seed_type = 'segwit'
+            elif ks.get('type') == 'old':
+                seed_type = 'old'
+            if seed_type is not None:
+                ks['seed_type'] = seed_type
+        self.data['seed_version'] = 40
+
+    def _convert_version_41(self):
+        # this is a repeat of upgrade 39, to fix wallet backup files (see #7339)
+        if not self._is_upgrade_method_needed(40, 40):
+            return
+        imported_channel_backups = self.data.pop('channel_backups', {})
+        imported_channel_backups.update(self.data.get('imported_channel_backups', {}))
+        self.data['imported_channel_backups'] = imported_channel_backups
+        self.data['seed_version'] = 41
+
+    def _convert_version_42(self):
+        # in OnchainInvoice['outputs'], convert values from None to 0
+        if not self._is_upgrade_method_needed(41, 41):
+            return
+        PR_TYPE_ONCHAIN = 0
+        requests = self.data.get('payment_requests', {})
+        invoices = self.data.get('invoices', {})
+        for d in [invoices, requests]:
+            for key, item in list(d.items()):
+                if item['type'] == PR_TYPE_ONCHAIN:
+                    item['outputs'] = [(_type, addr, (val or 0))
+                                       for _type, addr, val in item['outputs']]
+        self.data['seed_version'] = 42
+
+    def _convert_version_43(self):
+        if not self._is_upgrade_method_needed(42, 42):
+            return
+        channels = self.data.pop('channels', {})
+        for k, c in channels.items():
+            log = c['log']
+            c['fail_htlc_reasons'] = log.pop('fail_htlc_reasons', {})
+            c['unfulfilled_htlcs'] = log.pop('unfulfilled_htlcs', {})
+            log["1"]['unacked_updates'] = log.pop('unacked_local_updates2', {})
+        self.data['channels'] = channels
+        self.data['seed_version'] = 43
+
+    def _convert_version_44(self):
+        if not self._is_upgrade_method_needed(43, 43):
+            return
+        channels = self.data.get('channels', {})
+        for key, item in channels.items():
+            if bool(item.get('static_remotekey_enabled')):
+                channel_type = ChannelType.OPTION_STATIC_REMOTEKEY
+            else:
+                channel_type = ChannelType(0)
+            item.pop('static_remotekey_enabled', None)
+            item['channel_type'] = channel_type
+        self.data['seed_version'] = 44
+
+    def _convert_version_45(self):
+        from .lnaddr import lndecode
+        if not self._is_upgrade_method_needed(44, 44):
+            return
+        swaps = self.data.get('submarine_swaps', {})
+        for key, item in swaps.items():
+            item['receive_address'] = None
+        # note: we set height to zero
+        # the new key for all requests is a wallet address, not done here
+        for name in ['invoices', 'payment_requests']:
+            invoices = self.data.get(name, {})
+            for key, item in invoices.items():
+                is_lightning = item['type'] == 2
+                lightning_invoice = item['invoice'] if is_lightning else None
+                outputs = item['outputs'] if not is_lightning else None
+                bip70 = item['bip70'] if not is_lightning else None
+                if is_lightning:
+                    lnaddr = lndecode(item['invoice'])
+                    amount_msat = lnaddr.get_amount_msat()
+                    timestamp = lnaddr.date
+                    exp_delay = lnaddr.get_expiry()
+                    message = lnaddr.get_description()
+                    height = 0
+                else:
+                    amount_sat = item['amount_sat']
+                    amount_msat = amount_sat * 1000 if amount_sat not in [None, '!'] else amount_sat
+                    message = item['message']
+                    timestamp = item['time']
+                    exp_delay = item['exp']
+                    height = item['height']
+
+                invoices[key] = {
+                    'amount_msat':amount_msat,
+                    'message':message,
+                    'time':timestamp,
+                    'exp':exp_delay,
+                    'height':height,
+                    'outputs':outputs,
+                    'bip70':bip70,
+                    'lightning_invoice':lightning_invoice,
+                }
+        self.data['seed_version'] = 45
+
+    def _convert_invoices_keys(self, invoices):
+        # recalc keys of outgoing on-chain invoices
+        from .crypto import sha256d
+        def get_id_from_onchain_outputs(raw_outputs, timestamp):
+            outputs = [PartialTxOutput.from_legacy_tuple(*output) for output in raw_outputs]
+            outputs_str = "\n".join(f"{txout.scriptpubkey.hex()}, {txout.value}" for txout in outputs)
+            return sha256d(outputs_str + "%d" % timestamp).hex()[0:10]
+        for key, item in list(invoices.items()):
+            is_lightning = item['lightning_invoice'] is not None
+            if is_lightning:
+                continue
+            outputs_raw = item['outputs']
+            assert outputs_raw, outputs_raw
+            timestamp = item['time']
+            newkey = get_id_from_onchain_outputs(outputs_raw, timestamp)
+            if newkey != key:
+                invoices[newkey] = item
+                del invoices[key]
+
+    def _convert_version_46(self):
+        if not self._is_upgrade_method_needed(45, 45):
+            return
+        invoices = self.data.get('invoices', {})
+        self._convert_invoices_keys(invoices)
+        self.data['seed_version'] = 46
+
+    def _convert_version_47(self):
+        from .lnaddr import lndecode
+        if not self._is_upgrade_method_needed(46, 46):
+            return
+        # recalc keys of requests
+        requests = self.data.get('payment_requests', {})
+        for key, item in list(requests.items()):
+            lnaddr = item.get('lightning_invoice')
+            if lnaddr:
+                lnaddr = lndecode(lnaddr)
+                rhash = lnaddr.paymenthash.hex()
+                if key != rhash:
+                    requests[rhash] = item
+                    del requests[key]
+        self.data['seed_version'] = 47
+
+    def _convert_version_48(self):
+        # fix possible corruption of invoice amounts, see #7774
+        if not self._is_upgrade_method_needed(47, 47):
+            return
+        invoices = self.data.get('invoices', {})
+        for key, item in list(invoices.items()):
+            if item['amount_msat'] == 1000 * "!":
+                item['amount_msat'] = "!"
+        self.data['seed_version'] = 48
+
+    def _convert_version_49(self):
+        if not self._is_upgrade_method_needed(48, 48):
+            return
+        channels = self.data.get('channels', {})
+        legacy_chans = [chan_dict for chan_dict in channels.values()
+                        if chan_dict['channel_type'] == ChannelType.OPTION_LEGACY_CHANNEL]
+        if legacy_chans:
+            raise WalletFileException(
+                f"This wallet contains {len(legacy_chans)} lightning channels of type 'LEGACY'. "
+                f"These channels were created using unreleased development versions of Electrum "
+                f"before the first lightning-capable release of 4.0, and are not supported anymore. "
+                f"Please use Electrum 4.3.0 to open this wallet, close the channels, "
+                f"and delete them from the wallet."
+            )
+        self.data['seed_version'] = 49
+
+    def _convert_version_50(self):
+        if not self._is_upgrade_method_needed(49, 49):
+            return
+        requests = self.data.get('payment_requests', {})
+        self._convert_invoices_keys(requests)
+        self.data['seed_version'] = 50
+
+    def _convert_version_51(self):
+        from .lnaddr import lndecode
+        if not self._is_upgrade_method_needed(50, 50):
+            return
+        requests = self.data.get('payment_requests', {})
+        for key, item in list(requests.items()):
+            lightning_invoice = item.pop('lightning_invoice')
+            if lightning_invoice is None:
+                payment_hash = None
+            else:
+                lnaddr = lndecode(lightning_invoice)
+                payment_hash = lnaddr.paymenthash.hex()
+            item['payment_hash'] = payment_hash
+        self.data['seed_version'] = 51
+
     def _convert_imported(self):
         if not self._is_upgrade_method_needed(0, 13):
             return
@@ -788,6 +1094,19 @@ class WalletDB(JsonDB):
                 msg += "\nPlease open this file with Electrum 1.9.8, and move your coins to a new wallet."
         raise WalletFileException(msg)
 
+    def _add_db_creation_metadata(self):
+        # store this for debugging purposes
+        v = DBMetadata(
+            creation_timestamp=int(time.time()),
+            first_electrum_version_used=ELECTRUM_VERSION,
+        )
+        assert self.get("db_metadata", None) is None
+        self.put("db_metadata", v)
+
+    def get_db_metadata(self) -> Optional[DBMetadata]:
+        # field only present for wallet files created with ver 4.4.0 or later
+        return self.get("db_metadata")
+
     @locked
     def get_txi_addresses(self, tx_hash: str) -> List[str]:
         """Returns list of is_mine addresses that appear as inputs in tx."""
@@ -926,6 +1245,8 @@ class WalletDB(JsonDB):
         assert isinstance(tx_hash, str)
         assert isinstance(tx, Transaction), tx
         # note that tx might be a PartialTransaction
+        # serialize and de-serialize tx now. this might e.g. convert a complete PartialTx to a Tx
+        tx = tx_from_any(str(tx))
         if not tx_hash:
             raise Exception("trying to add tx to db without txid")
         if tx_hash != tx.txid():
@@ -1065,14 +1386,6 @@ class WalletDB(JsonDB):
         assert isinstance(txid, str)
         self.tx_fees.pop(txid, None)
 
-    @locked
-    def get_dict(self, name) -> dict:
-        # Warning: interacts un-intuitively with 'put': certain parts
-        # of 'data' will have pointers saved as separate variables.
-        if name not in self.data:
-            self.data[name] = {}
-        return self.data[name]
-
     @locked
     def num_change_addresses(self) -> int:
         return len(self.change_addresses)
@@ -1194,17 +1507,19 @@ class WalletDB(JsonDB):
             # note: for performance, "deserialize=False" so that we will deserialize these on-demand
             v = dict((k, tx_from_any(x, deserialize=False)) for k, x in v.items())
         if key == 'invoices':
-            v = dict((k, Invoice.from_json(x)) for k, x in v.items())
+            v = dict((k, Invoice(**x)) for k, x in v.items())
         if key == 'payment_requests':
-            v = dict((k, Invoice.from_json(x)) for k, x in v.items())
+            v = dict((k, Request(**x)) for k, x in v.items())
         elif key == 'adds':
             v = dict((k, UpdateAddHtlc.from_tuple(*x)) for k, x in v.items())
         elif key == 'fee_updates':
             v = dict((k, FeeUpdate(**x)) for k, x in v.items())
         elif key == 'submarine_swaps':
             v = dict((k, SwapData(**x)) for k, x in v.items())
-        elif key == 'channel_backups':
-            v = dict((k, ChannelBackupStorage(**x)) for k, x in v.items())
+        elif key == 'imported_channel_backups':
+            v = dict((k, ImportedChannelBackupStorage(**x)) for k, x in v.items())
+        elif key == 'onchain_channel_backups':
+            v = dict((k, OnchainChannelBackupStorage(**x)) for k, x in v.items())
         elif key == 'tx_fees':
             v = dict((k, TxFeesValue(*x)) for k, x in v.items())
         elif key == 'prevouts_by_scripthash':
@@ -1213,6 +1528,15 @@ class WalletDB(JsonDB):
             v = dict((k, ShachainElement(bfh(x[0]), int(x[1]))) for k, x in v.items())
         elif key == 'data_loss_protect_remote_pcp':
             v = dict((k, bfh(x)) for k, x in v.items())
+        # convert htlc_id keys to int
+        if key in ['adds', 'locked_in', 'settles', 'fails', 'fee_updates', 'buckets',
+                   'unacked_updates', 'unfulfilled_htlcs', 'fail_htlc_reasons', 'onion_keys']:
+            v = dict((int(k), x) for k, x in v.items())
+        # convert keys to HTLCOwner
+        if key == 'log' or (path and path[-1] in ['locked_in', 'fails', 'settles']):
+            if "1" in v:
+                v[LOCAL] = v.pop("1")
+                v[REMOTE] = v.pop("-1")
         return v
 
     def _convert_value(self, path, key, v):
@@ -1224,6 +1548,10 @@ class WalletDB(JsonDB):
             v = ChannelConstraints(**v)
         elif key == 'funding_outpoint':
             v = Outpoint(**v)
+        elif key == 'channel_type':
+            v = ChannelType(v)
+        elif key == 'db_metadata':
+            v = DBMetadata(**v)
         return v
 
     def _should_convert_to_stored_dict(self, key) -> bool:
@@ -1238,13 +1566,15 @@ class WalletDB(JsonDB):
         with self.lock:
             self._write(storage)
 
+    @profiler
     def _write(self, storage: 'WalletStorage'):
-        if threading.currentThread().isDaemon():
+        if threading.current_thread().daemon:
             self.logger.warning('daemon thread cannot write db')
             return
         if not self.modified():
             return
-        storage.write(self.dump())
+        json_str = self.dump(human_readable=not storage.is_encrypted())
+        storage.write(json_str)
         self.set_modified(False)
 
     def is_ready_to_be_used_by_wallet(self):
